#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#include <map>
#include <algorithm>

#include "util.h"
#include "gl_util.h"
#include "consts.h"
#include "globals.h"
#include "io.h"
#include "Image.h"

using namespace std;
using namespace globals;
using namespace consts;
using namespace io;

void draw_all_objects() {
   glActiveTexture(GL_TEXTURE0);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, shaders::bump_tex_handle);

   glActiveTexture(GL_TEXTURE1);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, shaders::decal_tex_handle);
   activate_program(shaders::prog);

   int bi_loc = getAttribLoc(shaders::prog, "binormal");
   int tan_loc = getAttribLoc(shaders::prog, "tangent");
   
   // Inform the shaders of the texture bindings done above 
   //  (ie, GL_TEXTURE0 and GL_TEXTURE1)
   glUniform1i(getUniLoc(shaders::prog, "bumpmap"), 0); 
   glUniform1i(getUniLoc(shaders::prog, "decal"), 1);

   glPushMatrix();
   glBegin(GL_QUADS); {
      glNormal3f(0, 0, 1);

      glTexCoord2f(0, 0);
      glVertexAttrib3f(bi_loc, 1, 0, 0);
      glVertexAttrib3f(tan_loc, 0, 1, 0);
      glVertex3f(-2.5f, 0, -2);

      glTexCoord2f(0, 1);
      glVertexAttrib3f(bi_loc, 1, 0, 0);
      glVertexAttrib3f(tan_loc, 0, 1, 0);
      glVertex3f(-2.5f, 5, -2);

      glTexCoord2f(1, 1);
      glVertexAttrib3f(bi_loc, 1, 0, 0);
      glVertexAttrib3f(tan_loc, 0, 1, 0);
      glVertex3f(2.5f, 5, -2);

      glTexCoord2f(1, 0);
      glVertexAttrib3f(bi_loc, 1, 0, 0);
      glVertexAttrib3f(tan_loc, 0, 1, 0);
      glVertex3f(2.5f, 0, -2);
   }
   glEnd();
   glPopMatrix();

   deactivate_shaders();
   glActiveTexture(GL_TEXTURE0);
   glDisable(GL_TEXTURE_2D);

   glActiveTexture(GL_TEXTURE1);
   glDisable(GL_TEXTURE_2D);
}

void setup_shaders() {
   // Load the 2 shaders, and link them into a program
   shaders::vshader = compile_vertex_shader("vert.glsl");
   shaders::fshader = compile_fragment_shader("frag.glsl");
   shaders::prog = compile_program(shaders::vshader, shaders::fshader);

   // Set up associated variables and textures
   Image* bumpmap_tex = Image::ReadBitmap("bumpmap.bmp");
   glGenTextures(1, &shaders::bump_tex_handle);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, shaders::bump_tex_handle);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, bumpmap_tex->width, bumpmap_tex->height, 0, GL_RGB, GL_UNSIGNED_BYTE, bumpmap_tex->data);

   Image* decal_tex = Image::ReadBitmap("decal.bmp");
   glGenTextures(1, &shaders::decal_tex_handle);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, shaders::decal_tex_handle);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, decal_tex->width, decal_tex->height, 0, GL_RGB, GL_UNSIGNED_BYTE, decal_tex->data);
}

void main_draw() {
   glutSetWindow(main_win::handle);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glPushMatrix(); {
      glRotatef(360.0f - (cam::ang_x * (180.0f / M_PI)), 1, 0, 0);
      glRotatef(360.0f - (cam::ang_y * (180.0f / M_PI)), 0, 1, 0);
      glTranslatef(-cam::pos_x, -cam::pos_y, -cam::pos_z);

      glLightfv(GL_LIGHT0, GL_POSITION, light::pos);

      // Draw Grid
      draw_base_grid();

      glEnable(GL_LIGHTING);
      glPushMatrix();
      draw_all_objects();
      glPopMatrix();
      glDisable(GL_LIGHTING);

      glPushMatrix();
      glColor3f(1, 1, 0);
      glTranslatef(light::pos[0], light::pos[1], light::pos[2]);
      glutSolidSphere(0.25, 10, 10);
      glPopMatrix();

   }
   glPopMatrix();
   glutSwapBuffers();
}

void update_camera(int dt) {
   if (warp_mouse)
      glutWarpPointer(main_win::width >> 1, main_win::height >> 1);

   float x = 0.0f;
   float y = 0.0f;
   float z = 0.0f;

   if (keys[FORWARD_KEY]) {
      x += -MOVE_DELTA * sin(cam::ang_y);
      z += -MOVE_DELTA * cos(cam::ang_y);
   }

   if (keys[BACKWARD_KEY]) {
      x += MOVE_DELTA * sin(cam::ang_y);
      z += MOVE_DELTA * cos(cam::ang_y);
   }

   if (keys[LEFT_KEY]) {
      x += -MOVE_DELTA * cos(cam::ang_y);
      z += MOVE_DELTA * sin(cam::ang_y);
   }

   if (keys[RIGHT_KEY]) {
      x += MOVE_DELTA * cos(cam::ang_y);
      z += -MOVE_DELTA * sin(cam::ang_y);
   }

   if (keys[RISE_KEY])
      y += MOVE_DELTA;

   if (keys[FALL_KEY])
      y += -MOVE_DELTA;

   cam::pos_x += x * dt;
   cam::pos_y += y * dt;
   cam::pos_z += z * dt;

   if (cam::pos_y < 0.0f)
      cam::pos_y = 0.0f;
}

void update(int val) {
   // Elapsed Time
   static int then;
   int now = glutGet(GLUT_ELAPSED_TIME);
   int dt = then - now;
   elapsed_time += dt;

   update_camera(dt);

   light::pos[0] = 2.5 * sin(elapsed_time / 2000.0f);
   light::pos[1] = 2.5 + 2.5 * cos(elapsed_time / 2000.0f);

   then = glutGet(GLUT_ELAPSED_TIME);

   glutSetWindow(main_win::handle);
   glutPostRedisplay();
   glutTimerFunc(10, update, val);
}

void create_main_window() {
   main_win::handle = glutCreateWindow(APP_TITLE);

   // Keyboard Callbacks
   glutKeyboardFunc(key_down);
   glutKeyboardUpFunc(key_up);
   glutSpecialFunc(special_key_down);
   glutSpecialUpFunc(special_key_up);

   // Mouse Callbacks
   glutMouseFunc(mouse_click);
   glutMotionFunc(io::mouse_motion);
   glutPassiveMotionFunc(io::mouse_motion);
   glutSetCursor(GLUT_CURSOR_NONE);
   glutEntryFunc(io::mouse_window);

   // Display callbacks
   glutReshapeFunc(glut_resize);
   glutDisplayFunc(main_draw);

   init_gl();
   glut_resize(main_win::width, main_win::height);
}

int main(int argc, char** argv) {
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

   glutInitWindowSize(main_win::width, main_win::height);
   glutInitWindowPosition(main_win::pos_x, main_win::pos_y);

   create_main_window();
   setup_shaders();

   glutTimerFunc(10, update, 0);

   glutMainLoop();

   return 0;
}
